#define DIMENSIONS 2

#include "spatial_parameters.hh"
#include "grid.hh"
#include <limits>

// A functor implementation of the gaussian function
class Gaussian{
private:
  double sigma;
  Vector x0;

public:
  Gaussian(double _sigma, Vector _x0)
    : sigma(_sigma), x0(_x0)
  {}
    
  // evalutate at x
  inline double operator()(const Vector & x) const
  {
    return 1.0/2.0/M_PI/sigma * exp(-x.two_norm_sqr()/2.0/sigma/sigma);
  }

  // evalutate derivative in d at x
  inline double derivative(const unsigned int d, const Vector & x) const
  {
    return -x[d]/sigma/sigma * (*this)(x);
  }
};

int main()
{
  // Setup grid.
  Vector extent( 2, 1.0 );           // 2d-domain: 1.0m x 1.0m
  std::vector<size_t> size( 2, 40 ); // 2d-grid: 40 x 40 cells = pixels resolution of our "startimage.tif"
  Grid grid( extent, size );         // Create the grid.

  // Input image.
  cout << endl;
  cout << "Create SpatialParameters object for the original image." << endl;
  SpatialParameters input_image( "startimage.tif", grid, "image.cfg", "Values" );
  
  // Output image.
  cout << endl;
  cout << "Create SpatialParameters object for the output image." << endl;
  SpatialParameters output_image( "startimage.tif", grid, "image.cfg", "Values" );
  

  double max_result = 0.0;
  double min_result = 1000.0;

  // Begin outer iteration.
  cout << endl;

  std::cout << "Loop over the grid and calculate the value of the convolution integral for each cell!" << std::endl;

  // setup clock
  clock_t timer_begin = clock();

  for( ElementIterator git = grid.begin()
	 ; git != grid.end()
	 ; ++git)
    {
      // Setup appropriate gaussian.
      Vector x0 = (*git).barycenter();
      const double sigma = 0.025;
      const Gaussian g( sigma, x0 );

      // Reset image value.
      double & result = output_image( git.id() );
      result = 0;
  
      // Inner loop over the grid (This is to calculate the convolution integral for each point x0).
      for( ElementIterator it = grid.begin()
	     ; it != grid.end()
	     ; ++it )
	{
	  Element & element = *it;
	  const Vector & center = element.barycenter();
	  const size_t element_id = it.id();
	  const double image_value = input_image( element_id );  // = f_i
	  
	  if( image_value > 1e-12 )
	    {
	      // Iterate over faces to evaluate the line integral on each cell
	      for( FaceIterator fit = (*it).begin()
		     ; fit != (*it).end()
		     ; ++fit )
		{
		  Face & face = *fit;
		  const double & area = face.area();   // in 2d, this is actually the length of an edge		  		  
		  // In 2d(3d), the faces in each cell are numbered in the following way:
		  // face_id_ = 0=west, 1=east, 2=south, 3=north (, 4=bottom, 5=top )
		  const unsigned int normal_component = face.normal_component();
		  const int normal_direction = face.normal_direction();
				  				  
		  // Compute x - x0 at face center
		  Vector facecenter = center;
		  facecenter[ normal_component ] += ((double) normal_direction) * face.dist_to_barycenter();		  
		  for( size_t i=0; i<facecenter.size(); ++i )
		    facecenter[i] -= x0[i];
				  				  
		  // Compute normal * nabla Gaussian
		  const double n_nabla_g = - M_PI*sigma*sigma*sigma*
		    ((double) normal_direction) * g.derivative( normal_component, facecenter );
				  
		  // Contribution to inner integral.
		  result += n_nabla_g * image_value * area;
				  
		} // face iterator
	    } // end-if( image_value > 1e-12 )
	} // Element Iterator for the convolution

      max_result = max( max_result, result );
      min_result = min( min_result, result );
	  
    } // Element Iterator for the grid

  clock_t timer_end = clock();
  cout << "Time for convolution integral: " << (double)( timer_end - timer_begin ) / CLOCKS_PER_SEC << endl;

  cout << "minimal value = " << min_result << endl;
  cout << "maximal value = " << max_result << endl;

  output_image.save_as_image("convolution.tif");

  return 0;
}
