#define DIMENSIONS 2

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <limits>
#include <libconfig.h++>

#include "CImg.h"
#include "Vector.hh"

using namespace std;
using namespace cimg_library;
using namespace libconfig;


// A functor implementation of the gaussian function
class Gaussian{
private:
  double sigma;
  Vector x0;
  
public:
  Gaussian( double _sigma, Vector _x0 )
    : sigma(_sigma), x0(_x0)
  {}
    
  // evaluate at x
  inline double operator()(const Vector & x) const
  {
    return 1.0/2.0/M_PI/sigma * exp(-x.two_norm_sqr()/2.0/sigma/sigma);
  }
  
  // evaluate derivative in d at x
  inline double derivative(const unsigned int d, const Vector & x) const
  {
    return -x[d]/sigma/sigma * (*this)(x);
  }

  // evaluate laplacian
  inline double laplacian(const Vector & x) const
  {
    return ( x.two_norm_sqr()/sigma/sigma - 2.0 ) * (*this)(x) /sigma/sigma;
  }

};

int main()
{
  const double sigma = 0.025;

  vector<double> mapping_vector( 256 ); // initialize vector of length 256, all values = 0.0 by default
  

  // Load cfg file.
  Config cfg;
  string _map_file = string("image.cfg");
  try
    {
      string mapkey = string("Values.Map");
      // Read map from .cfg file
      cout << "Reading file " << _map_file.c_str() << endl;
      cfg.readFile( _map_file.c_str() );

      if(!cfg.exists(  mapkey.c_str() ))
	{
	  cerr << "The given name of the map \"" << mapkey.c_str() 
	       << "\" was not found in the configuration file" << endl;
	  exit(1);
	}
      cout << "Found map " << mapkey.c_str() << endl;
	  
      Setting & mappings = cfg.lookup(mapkey);
      const int n_mappings = mappings.getLength();
      if(n_mappings < 1)
	{
	  cerr << "Error: Map was empty." << endl;
	  exit(1);
	}
      
      cout << "Found " << n_mappings << " mappings:" << endl;

      for( int i=0; i<n_mappings; i++ )
	{
	  cout << int( mappings[i][0] ) << " ---> ";
	  for( int j=1; j<mappings[i].getLength(); j++ )
	    {
	      mapping_vector[ int( mappings[i][0] ) ] = double( mappings[i][j] );
	      cout << double( mappings[i][j] ) << "  ";
	    }
	  cout << endl;
	}
      
    }
  catch(ParseException & exc)
    {
      cerr << "Error reading config file "<< _map_file << ": " << endl;
      cerr << "\t " << exc.getError() <<" in line "<< exc.getLine() << endl;
      exit(1);
    }


  // Load the given image.

  string _image_filename = string("startimage.tif"); // Our image is a grey-scale 8-bit image.
  typedef unsigned char BYTE;   // This type has values from 0..255 (= 8bit)

  // Check whether this is a tiff file
  if( _image_filename.rfind(".tif") != string::npos )
    {
      cout << "Reading values from a tiff file" << endl;
      
      // Load tiff image
      CImg<BYTE> tmp_image( _image_filename.c_str() );

#ifdef ASCII_PLOT
      cout << "ascii plot of the tiff image, xyz = " 
	   << tmp_image.dimx() << "x" 
	   << tmp_image.dimy() << "x" 
	   << tmp_image.dimz() << endl << endl;
      for( size_t i=0; i<tmp_image.dimx()*tmp_image.dimy(); ++i )
	{
	  if( i % tmp_image.dimx() == 0 )
	    cout << endl;
	  if( 255 == static_cast<int>( tmp_image[ i ] ) )
	    cout << " ";
	  else
	    cout << "X";
	}
      cout << endl;
      cout << endl;
#endif

      // Calculate the convolution.	  

      double Lx = 1.0;
      double Ly = 1.0;
	  
      int Nx = tmp_image.dimx();
      int Ny = tmp_image.dimy();

      double dx = Lx / (double) Nx;
      double dy = Ly / (double) Ny;

      int nAllPixels = Nx * Ny;
	  
      cout << " Define the cell-centers x_k..." << endl;
      vector<Vector> x_k;
      x_k.resize( nAllPixels );

      // add your code here

      cout << "Define normal vectors for each of the four directions..." << endl;
      vector<Vector> normal;
      normal.resize( 4 );  // There are only for edges in a rectangular cell!

      // add your code here

      cout << "Define the four face-centers z_j for each cell..." << endl;
      vector< vector< Vector > > z_j;
      z_j.resize( nAllPixels );
	  
      // add your code here

      CImg<double> new_image = CImg<double>( Nx, Ny, 1, 1 );
      vector<double> newdata( Nx*Ny );

      // setup clock
      clock_t timer_begin = clock();

	  	  
      for( int k=0; k<nAllPixels; k++ )
	{
	  const Gaussian gaussian( sigma, x_k[k] );
		  
	  double conv_k = 0.0;
	  for( int i=0; i<nAllPixels; i++ )
	    {
	      double f_ii = mapping_vector[ static_cast<int>(tmp_image[ i ]) ];
	      if( f_ii > 1e-12 )
		{
		  // add your code here
		  conv_k += //...
		}
			  
	    } // end i-loop

	  newdata[ k ] = conv_k;

	} // end k-loop



      clock_t timer_end = clock();
      cout << "Time for convolution integral: " << (double)( timer_end - timer_begin ) / CLOCKS_PER_SEC << endl;


      double min = std::numeric_limits<double>::max();
      double max = std::numeric_limits<double>::min();
      for(size_t i=0; i<newdata.size(); ++i)
	{
	  if(min > newdata[i])
	    min = newdata[i];
	  if(max < newdata[i])
	    max = newdata[i];
	}

      cout << "min=" << min << endl;
      cout << "max=" << max << endl;
  
      // color palette: 0..255
      double min_value = 1000;
      double max_value = -1000;
	  
      for(size_t i=0; i<newdata.size(); ++i)
	{
	  // Rescaling values to adjust to color palette values!
	  double value = ( newdata[i] - min ) / ( max - min ) * 255.0;
	  new_image[ i ] = value;
		  min_value = std::min( value, min_value );
		  max_value = std::max( value, max_value );
		}
	  
	  cout << "min_color=" << min_value << endl;
	  cout << "max_color=" << max_value << endl;
	  
	  // Save new image.
	  new_image.save("fast.tif");
	  
  }

  return 0;
}
