/*
   This is an implementation by Dan Popovic of the structured
   rectangular grid as requested for exercise 2 of the lecture
   Numerics of Transport in Porous Media.

   However, to avoid the need to deal with template classes it defines
   the types: Grid, ElementIterator, Element, FaceIterator, Face as
   described in exercise 2. Therefore, it requires the precompiler
   macro DIMENSIONS to be set. Hence, make sure to have a line like

   #define DIMENSIONS 2 

   before the line

   #include "grid.hh"

   to avoid compiler errors because of an undefined symbol "DIMENSIONS".

   This implementation requires the class Vector, the header file
   "Vector.hh" must be included.

*/

#ifndef __FV_GRID_HH_
#define __FV_GRID_HH_

#include <iostream>
#include <cstdlib>
#include <vector>
#include "Vector.hh"

namespace FiniteVolumeGrid{

  // forward declarations
  template<class T,int dim> class FaceIterator;
  template<class T,int dim> class ElementIterator;

  //==============================================================================
  // face class
  //==============================================================================
  template<class T,int dim>
  class CFace
  {
  public:
    friend class FaceIterator<T,dim>;

    // constructors
    CFace();
    CFace( const std::vector<size_t>& nc
		   , const std::vector<size_t>& df
		   , const Vector&e
		   , const size_t& id
		   , size_t eid );

    // access to face
    inline const T & area() const;
    inline const T & dist_to_barycenter() const;
    inline const bool & is_boundary() const;
    inline unsigned int normal_component() const;
    inline int normal_direction() const;

  private:
    // face data
    size_t face_id_;                            // id of face (2d: 0=west, 1=east, 2=south, 3=north, 3d: 0..5)
    size_t element_id_;                    // id of father element
    int out_id_;                                 // id of neighbor element
    T area_;                                     // area of face
    bool is_boundary_;                           // boundary flag
    T distance_to_barycenter;			 // distance to barycenter

    // grid info references
    const std::vector<size_t>& ncells_;    // number of cells
    const std::vector<size_t>& dimfactor_; // dimension factors
    const Vector& element_extent_;       // extent of father element

    // set methods
    inline void set_boundary_flag();
    inline void set_id_neighbor();
    inline void set_area();
    inline void set_distance();
  };

  // constructors
  template<class T,int dim>
  CFace<T,dim>::CFace() 
	: face_id_(0)
	, element_id_(0)
	, out_id_(-1)
	, area_(0.0)
	, is_boundary_(false)
	, ncells_(0)
	, dimfactor_(0)
	, element_extent_(0)
  {
  }

  template<class T,int dim>
  CFace<T,dim>::CFace( 
					  const std::vector<size_t>& nc
					  , const std::vector<size_t>& df
					  , const Vector& e
					  , const size_t& face_id
					  , size_t eid
					   ) 
	: face_id_( face_id )
	, element_id_(eid)
	, out_id_(-1)
	, ncells_(nc)
	, dimfactor_(df)
	, element_extent_(e)
  {
    set_boundary_flag();
    set_area();
    set_id_neighbor();
    set_distance();
  }

  // set boundary flag
  template<class T,int dim>
  inline void CFace<T,dim>::set_boundary_flag()
  {
    is_boundary_ = false;

    // project current position onto all directions to get number of cells in each direction
    size_t eid = element_id_;
    std::vector<size_t> n(dim+1, 0);
    for (int i=dim; i>=0; --i)
      {
        n[i] = eid / dimfactor_[i] + 1;
        eid -= (n[i]-1) * dimfactor_[i];
      }

    // now set boundary flag
    const unsigned int norm_comp = normal_component();
    if ( (n[norm_comp]-1==0 && normal_direction()<0) || 
	 (n[norm_comp]==ncells_[norm_comp] && normal_direction()>0) ) 
      is_boundary_ = true;

    return;
  }

  // set id of neighbouring element
  template<class T,int dim>
  inline void CFace<T,dim>::set_id_neighbor()
  {
    out_id_ = -1;
    if (!is_boundary_) out_id_ = element_id_ + normal_direction() * dimfactor_[normal_component()];

    return;
  }

  // set id of neighbouring element
  template<class T,int dim>
  inline void CFace<T,dim>::set_distance()
  {
    distance_to_barycenter = 0.0;
    //if (!is_boundary_) 
	distance_to_barycenter = element_extent_[normal_component()]*0.5;

    return;
  }

  // calculate area of face
  template<class T,int dim>
  inline void CFace<T,dim>::set_area()
  {
    area_ = 1.0;
    for( int j=0; j<dim+1; ++j )
      area_ *= element_extent_[j];
    area_ /= element_extent_[normal_component()];

    return;
  }

  // area of face
  template<class T,int dim>
  inline const T & CFace<T,dim>::area() const
  {
    return area_;
  }

  // distance to barycenter of neighbor element
  template<class T,int dim>
  inline const T & CFace<T,dim>::dist_to_barycenter() const
  {
    return distance_to_barycenter;
  }

  // is face on boundary?
  template<class T,int dim>
  inline const bool & CFace<T,dim>::is_boundary() const
  {
    return is_boundary_;
  }

  // as we have 2*d faces in the d-dimensional case the normal component is easy:
  template<class T,int dim>
  inline unsigned int CFace<T,dim>::normal_component() const
  {
    return face_id_ / 2;
  }

  // as we have 2*d faces in the d-dimensional case the normal direction is also easy:
  template<class T,int dim>
  inline int CFace<T,dim>::normal_direction() const
  {
    // result of id_%2 is in [0,1] --> transform it to [-1,1]
    return 2 * (face_id_ % 2) - 1;
  }

  //==============================================================================
  // face iterator class
  //==============================================================================
  template<class T,int dim>
  class FaceIterator
  {
  private:
    CFace<T,dim> face_; // the encapsulated face
	
  public:
    // constructors
    FaceIterator();
    FaceIterator( const std::vector<size_t>& nc
				  , const Vector& e
				  , const std::vector<size_t>& dimfactor
				  , const size_t& eid
				  , const size_t& fid );

    // operators
    CFace<T,dim>& operator*();
    FaceIterator& operator++();
    bool operator!=(const FaceIterator& fit) const;

    // id of inner and outer element
    inline size_t in_id() const;
    inline size_t out_id() const;

    // reset iterator
    FaceIterator& reset();

  };

  // constructors
  template<class T,int dim>
  FaceIterator<T,dim>::FaceIterator() : face_()
  {
  }

  template<class T,int dim>
  FaceIterator<T,dim>::FaceIterator( const std::vector<size_t>& nc
									 , const Vector& e
									 , const std::vector<size_t>& dimfactor
									 , const size_t& eid
									 , const size_t& fid ) 
	: face_( nc
			 , dimfactor
			 , e
			 , fid
			 , eid )
  {
  }

  // dereferencing
  template<class T,int dim>
  CFace<T,dim>& FaceIterator<T,dim>::operator*()
  {
    return face_;
  }

  // increment
  template<class T,int dim>
  FaceIterator<T,dim>& FaceIterator<T,dim>::operator++()
  {
    (face_.face_id_)++;
    face_.set_boundary_flag();
    face_.set_area();
    face_.set_id_neighbor();
    face_.set_distance();

    return *this;
  }

  // comparison
  template<class T,int dim>
  bool FaceIterator<T,dim>::operator!=(const FaceIterator<T,dim>& fit) const
  {
    return ( ((*this).face_.element_id_ == fit.face_.element_id_) && ((*this).face_.face_id_ != fit.face_.face_id_) );
  }

  // reset
  template<class T,int dim>
  FaceIterator<T,dim>& FaceIterator<T,dim>::reset()
  {
    face_.id_ = 0;
    face_.set_boundary_flag();
    face_.set_area();
    face_.set_id_neighbor();
    face_.set_distance();

    return *this;
  }

  // methods returning element id's
  template<class T,int dim>
  inline size_t FaceIterator<T,dim>::in_id() const
  {
    return face_.element_id_;
  }

  template<class T,int dim>
  inline size_t FaceIterator<T,dim>::out_id() const
  {
    return face_.out_id_;
  }

  //==============================================================================
  // element class
  //==============================================================================
  template<class T,int dim>
  class CElement
  {
    friend class ElementIterator<T,dim>;
	
  public:
    typedef FaceIterator<T,dim-1> FaceIter;

    // constructors
    CElement();
    CElement(
			const Vector& gridextent
			, const std::vector<size_t>& nc
			, const std::vector<size_t>& dimfactor
			, size_t id
			);

    // access to element
    inline const T & volume() const;
    const Vector& barycenter() const;
    const Vector& extent() const;

    // face iterators
    FaceIter begin() const;
    FaceIter end() const;

  private:
    // element data
    size_t id_;
    T volume_;
    Vector barycenter_;
    Vector extent_;
    std::vector<size_t> ncells_;
    const std::vector<size_t>& dimfactor_;
  };

  // constructors
  template<class T,int dim>
  CElement<T,dim>::CElement() 
	: id_(-1)
	, volume_(0.0)
	, barycenter_(dim,0)
	, extent_(dim,0)
	, ncells_(0)
	, dimfactor_(dim,0)
  {
  }

  template<class T,int dim>
  CElement<T,dim>::CElement(
						  const Vector& gridextent
						  , const std::vector<size_t>& numcells
						  , const std::vector<size_t>& dimfactor
						  , size_t id 
						  )
    : 
	id_(id)
	, barycenter_(dim,0)
	, extent_(dim,0)
	, ncells_(numcells)
	, dimfactor_(dimfactor)
  {
    if ( gridextent.size() != dim )
      {
        std::cerr << "Size mismatch, dim>ncells[dim]" << std::endl;
        exit(EXIT_FAILURE);
      }

    // set extent, volume and barycenter of element
    volume_ = 1.0;
    std::vector<size_t> n(dim,0);
    for (int i=dim-1; i>=0; --i)
      {
        // local extent and volume
        extent_[i] = gridextent[i]/numcells[i];
        volume_ *= extent_[i];

        // barycenter
        n[i] = id / dimfactor[i] + 1;
        id -= (n[i]-1) * dimfactor[i];
        barycenter_[i] = (n[i] - 0.5) * extent_[i];
      }
  }

  // volume of element
  template<class T,int dim>
  inline const T & CElement<T,dim>::volume() const
  {
    return volume_;
  }

  // barycenter of element
  template<class T,int dim>
  const Vector& CElement<T,dim>::barycenter() const
  {
    return barycenter_;
  }

  // retrieve element extent
  template<class T,int dim>
  const Vector& CElement<T,dim>::extent() const
  {
    return extent_;
  }

  // return iterator to first face
  template<class T,int dim>
  typename CElement<T,dim>::FaceIter CElement<T,dim>::begin() const
  {
    return FaceIter( ncells_, extent_, dimfactor_, id_, 0 );
  }

  // return iterator to one behind last face
  template<class T,int dim>
  typename CElement<T,dim>::FaceIter CElement<T,dim>::end() const
  {
    return FaceIter(ncells_, extent_, dimfactor_, id_, 2*dim);
  }

  //==============================================================================
  // element iterator class
  //==============================================================================
  template<class T,int dim>
  class ElementIterator
  {
  public:
    // constructors
    ElementIterator();
    ElementIterator( const Vector& ext
					 , const std::vector<size_t>& ncells
					 , const std::vector<size_t>& dimfactor
					 , size_t id 
					 );

    // operators
    ElementIterator& operator++();
    bool operator!=(const ElementIterator& ei) const;
    CElement<T,dim>& operator*();

    size_t id() const;  // get id of current element
    ElementIterator& reset(); // reset iterator

  private:
    CElement<T,dim> element_;                     // the element encapsulated by iterator
    const std::vector<size_t>& dimfactor_; // reference to factors used in calculation of barycenter
  };

  // constructors
  template<class T,int dim>
  ElementIterator<T,dim>::ElementIterator() : element_(0), dimfactor_(0)
  {
  }

  template<class T,int dim>
  ElementIterator<T,dim>::ElementIterator(const Vector& ext, const std::vector<size_t>& ncells, 
					  const std::vector<size_t>& dimfactor, size_t id) :
    element_( ext, ncells, dimfactor, id ), dimfactor_(dimfactor)
  {
  }

  // reset iterator
  template<class T,int dim>
  ElementIterator<T,dim>& ElementIterator<T,dim>::reset()
  {
    element_.id_=0;

    // reset barycenter of element
    size_t id = element_.id_;
    std::vector<size_t> n(dim,0);
    for (int i=dim-1; i>=0; --i)
      {
        n[i] = id / dimfactor_[i] + 1;
        id -= (n[i]-1) * dimfactor_[i];
        element_.barycenter_[i] = (n[i] - 0.5) * element_.extent_[i];
      }

    return *this;
  }

  // increment uses friendship of Iterator to Element
  template<class T,int dim>
  ElementIterator<T,dim>& ElementIterator<T,dim>::operator++()
  {
    ++(element_.id_);

    // reset barycenter of element
    size_t id = element_.id_;
    std::vector<size_t> n(dim,0);
    for (int i=dim-1; i>=0; --i)
      {
        n[i] = id / dimfactor_[i] + 1;
        id -= (n[i]-1) * dimfactor_[i];
        element_.barycenter_[i] = (n[i] - 0.5) * element_.extent_[i];
      }

    return *this;
  }

  // dereferencing
  template<class T,int dim>
  CElement<T,dim>& ElementIterator<T,dim>::operator*()
  {
    return element_;
  }

  // comparison of two iterators
  template<class T,int dim>
  bool ElementIterator<T,dim>::operator!=(const ElementIterator& ei) const
  {
    return (ei.element_.id_ != this->element_.id_ );
  }

  // returns id of current element
  template<class T,int dim>
  size_t ElementIterator<T,dim>::id() const
  {
    return element_.id_;
  }

  //==============================================================================
  // grid class
  //==============================================================================
  template<class T,int dim>
  class FVGrid
  {
  public:
    enum { dimensions = dim };
    typedef T ctype;
    typedef ElementIterator<T,dim> ElementIter;
    typedef FaceIterator<T,dim-1> FaceIter;

    // constructors
    FVGrid(const Vector e, const std::vector<size_t> nc);

    // access to grid info
    const Vector& extent() const;
    size_t cells(unsigned int d) const;
    const std::vector<size_t> & cells() const;

    // iterators pointing to first and last element of grid
    ElementIter begin() const;
    ElementIter end() const;

    ElementIter get_element_iterator(const size_t id) const;

  private:
    // grid data
    Vector extent_;
    std::vector<size_t> ncells_;
    std::vector<size_t> dimfactor_;
    size_t idmax_;
  };

  template<class T,int dim>
  FVGrid<T,dim>::FVGrid( 
						const Vector e
						, const std::vector<size_t> nc 
						 ) 
	: 
	extent_(e)
	, ncells_(nc)
	, dimfactor_(dim,1)
  {
    // grids with dim > 3 seem to be complicated
    if ( dim<1 || dim>3 )
      {
        std::cerr << "Bad grid dimension!" << std::endl;
        exit(EXIT_FAILURE);
      }

    // check consistency
    if ( (e.size()!=dim) || (nc.size()!=dim) )
      {
        std::cerr << "Size mismatch, dim != size()" << std::endl;
        exit(EXIT_FAILURE);
      }

    // set max id of grid elements
    idmax_ = 1;
    for (int i=0; i<dim; ++i) idmax_ *= ncells_[i];
    --idmax_;

    // set dimension factors  : dimfactor_[2] = ncells_[0]*ncells_[1]; dimfactor_[1] = ncells_[0]; dimfactor_[0]=1;
    size_t fac = 1;
    for (int i=1; i<dim; ++i)
      {
        fac *= ncells_[i-1];
        dimfactor_[i] = fac;
      }
  }

  // retrieve grid info
  template<class T,int dim>
  const Vector& FVGrid<T,dim>::extent() const
  {
    return extent_;
  }

  template<class T,int dim>
  size_t FVGrid<T,dim>::cells(unsigned int d) const
  {
    if ( d > (ncells_.size()-1) )
      {
        throw("Size mismatch, dim>ncells[dim]");
        return ncells_[ncells_.size()-1];
      }
    return ncells_[d];
  }

  template<class T,int dim>
  const std::vector<size_t> & FVGrid<T,dim>::cells() const
  {
    return ncells_;
  }


  // iterators pointing to first and last element of grid
  template<class T,int dim>
  typename FVGrid<T,dim>::ElementIter FVGrid<T,dim>::begin() const
  {
    return ElementIter(extent_, ncells_, dimfactor_, 0);
  }

  template<class T,int dim>
  typename FVGrid<T,dim>::ElementIter FVGrid<T,dim>::end() const
  {
    return ElementIter(extent_, ncells_, dimfactor_, idmax_+1);
  }

  // iterators pointing to first and last element of grid
  template<class T,int dim>
  typename FVGrid<T,dim>::ElementIter FVGrid<T,dim>::get_element_iterator(const size_t id) const
  {
    return ElementIter(extent_, ncells_, dimfactor_, id);
  }

}; // namespace

typedef FiniteVolumeGrid::FVGrid<double,DIMENSIONS> Grid;
typedef FiniteVolumeGrid::ElementIterator<double,DIMENSIONS> ElementIterator;
typedef FiniteVolumeGrid::CElement<double,DIMENSIONS> Element;
typedef FiniteVolumeGrid::FaceIterator<double,DIMENSIONS-1> FaceIterator;
typedef FiniteVolumeGrid::CFace<double,DIMENSIONS-1> Face;
#endif
