/* This implements the class representing the conductivity field. It
   makes use of the LGPL library libconfig which has to be available
   in order to compile this header file. It also requires the CImg.h
   library with some kind of tiff support.

   Furthermore, it requires the precompiler macro DIMENSIONS to be
   defined.Hence, make sure to have a line like

      #define DIMENSIONS 2 

   before the line

      #include "grid.hh"

   to avoid compiler errors because of an undefined symbol "DIMENSIONS".

   This implementation requires the class Vector, the header file
   "Vector.hh" must be included.
   
 */

#ifndef SPATIAL_PARAMETERS_HH
#define SPATIAL_PARAMETERS_HH

#include "CImg.h"
#include "Vector.hh"
#include "grid.hh"
#include <libconfig.h++>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <limits>

using namespace std;
using namespace cimg_library;

//! \class SpatialParameters 
//!
//! Provides access to the scalar parameters field
class SpatialParameters
{
private:
  enum {dimensions = Grid::dimensions};
  typedef CImg<double> Image;
  vector<Image> image_vector;   // storage for a series of images
  Vector extent;
  Vector cell_length;
  const Grid & grid;
  int n_parameter_values;

  // This provides a simple rescaling of an image
  void rescale_image(const Image & in, Image & out)
  {
    size_t z_size = Grid::dimensions==3 ? grid.cells()[2] : 1;
	// Create an image with the dimensions of the grid, initializing all pixels with value 1
    out = Image( grid.cells()[0], grid.cells()[1], z_size, 1 );
	
    size_t id=0;
    const size_t XY = in.dimx()*in.dimy();
    const size_t X = in.dimx();
    size_t x_factor = out.dimx() / in.dimx();
    size_t y_factor = out.dimy() / in.dimy();
    size_t z_factor = out.dimz() / in.dimz();
    cout << "Scaling factors: "
		 << x_factor<<" "
		 << y_factor<<" "
		 << z_factor<<endl;
	
    if(x_factor==0)x_factor=1;
    if(y_factor==0)y_factor=1;
    if(z_factor==0)z_factor=1;
	
    for(size_t z=0; z< (Grid::dimensions == 3 ? grid.cells()[2] : 1); ++z){
      for(size_t y=0; y<grid.cells()[1]; ++y){
		for(size_t x=0; x<grid.cells()[0]; ++x){
		  const size_t tmp_id = 
			+  x / x_factor 
			+ (y / y_factor)*X
			+ (z / z_factor)*XY;
		  out[id++] = in[tmp_id];
		}
      }
    }
  }
  
public:

  //! Constructor reading parameter values from either a file in the
  //! tiff or native cimg format. Notice that the image will be
  //! mirrored top to bottom. The values read from the image may be
  //! mapped to different values according to a map defined in the
  //! file _map_file. As such a file may hold multiple maps the name
  //! of the map has to be given too. The size of the image may vary
  //! from the size of the grid as long as the grid size is an integer
  //! multiple of the image size.
  SpatialParameters(
		    string _image_filename
		    , const Grid & _grid
		    , string _map_file = string("")
		    , string _map_name = string("") 
		    )
    : 
    extent(_grid.extent())
    , cell_length(dimensions)
    , grid(_grid)
    , n_parameter_values(0)
  {
    
    // 
    vector<Image> small_image;

    // Check whether this is a tiff file
    if(_image_filename.rfind(".tif") != string::npos){
      
      cout << "Reading values from a tiff file" << endl;
      
      // Load tiff image
      CImg<short> tmp_image( _image_filename.c_str() );

      // This 256 x 2 matrix shall be filled with the original values in the 1st column and mapped values in the 2nd column.
      vector<vector<double> > mapping_matrix(256); 

      
      if(_map_file.length() <= 0){
		cerr << "No Map to convert byte values to floats was found." << endl;
		exit(1);
      }

      if(_map_name.length() <= 0){
		cerr << "The name of the map in the configuration file has to be defined." << endl;
		exit(1);
      }


      cout<< "Found a map to map tiff byte values to floating point numbers."<< endl;

      using namespace libconfig;

      string mapkey = _map_name.c_str() + string(".Map");


      //////////////////////////////////////////////////////////////////
      // Read map from file
      Config cfg;
      try
		{
		  // Read map from .cfg file
		  cout << "Reading file " << _map_file.c_str() << endl;
		  cfg.readFile( _map_file.c_str() );
		  
		  if(!cfg.exists(  mapkey.c_str() ))
			{
			  cerr << "The given name of the map \"" << mapkey.c_str() 
				   << "\" was not found in the configuration file" << endl;
			  exit(1);
			}
		  cout << "Found map " << mapkey.c_str() << endl;
		  
		  Setting & mappings = cfg.lookup(mapkey);
		  const int n_mappings = mappings.getLength();
		  if(n_mappings < 1)
			{
			  cerr << "Error: Map was empty." << endl;
			  exit(1);
			}
		  cout << "Found " << n_mappings << " mappings." << endl;
		  
		  for(int i=0; i<n_mappings; ++i)
			{
			  // cout << "i = " << i << endl;
			  int n_values = mappings[i].getLength();   // number of all values per mapping
			  // cout << "n_values = " << n_values << endl;
			  
			  // check consistency of .cfg data
			  if(n_parameter_values != 0 && n_values != n_parameter_values + 1)
				{
				  cerr << "Error: Every map entry must have the same size." << endl;
				  exit(1);
				}

			  // take over number of parameter values from the cfg file
			  n_parameter_values = n_values - 1;
			  
			  if(n_values < 2)
				{
				  cerr << "Error: Every map entry must be a list of at least size 2." << endl;
				  exit(1);
				}


			  int from = mappings[i][0];   // in our case: mappings[0][1] = 255, mappings[1][1] = 0
			  // cout << "from = " << from << endl;

			  mapping_matrix[ from ].resize( n_values );   // create 1+1=2 columns for this row
			  
			  mapping_matrix[ from ][ 0 ] = from;  // original value (left column), all other values are initialized to zero

			  /*
				cout << (int) mappings[ 0 ][ 0 ] << endl;
				cout << (int) mappings[ 1 ][ 0 ] << endl;
				cout << (double) mappings[ 0 ][ 1 ] << endl;
				cout << (double) mappings[ 1 ][ 1 ] << endl;
				cout << "done."  << endl;
			  */

			  cout << "Mapping color " << from << " to value "; 

			  for( int v=1; v<n_values; v++ )
				{
				  mapping_matrix[ from ][ v ] = (double) mappings[ i ][ v ];
				  cout << fixed << setw(5) << mapping_matrix[from][v] << " ";
				}
			  cout << endl;
			}
		}
	  
      catch(ParseException & exc)
		{
		  cerr << "Error reading config file "<< _map_file << ": " << endl;
		  cerr << "\t " << exc.getError() <<" in line "<< exc.getLine() << endl;
		  exit(1);
		}
      
      //////////////////////////////////////////////////////////////////
	  

      // Perform the mapping: color --> value
	  cout << "n_parameter_values = " << n_parameter_values << endl;
      small_image.resize( n_parameter_values );
      for( int p=0; p < n_parameter_values; ++p )
		{
		  // Construct new image with the same size as the loaded tiff image and initial value 1:
		  small_image[ p ] = Image( tmp_image.dimx(), tmp_image.dimy(), tmp_image.dimz(), 1 );
		  const size_t small_image_size = small_image[ p ].size();
		  for( size_t i=0; i<small_image_size; ++i )
			{
			  if( mapping_matrix[ tmp_image[i] ].size() == 0 )
				{
				  cerr << "Error. Found color in picture without a mapping target"<< endl;
				  exit(1);
				}
			  // The image pixels now get the values from the cfg file.
			  small_image[ p ][ i ] = mapping_matrix[ tmp_image[i] ][ p + 1 ];
			}
		}

    }//tiff

    // Check whether this is a cimg file
    else if( _image_filename.rfind(".cimg") != string::npos )
	  {
		n_parameter_values = 1;
		cout << "Reading values from a cimg file" << endl;
		small_image.resize( n_parameter_values );
		small_image[0] = Image( _image_filename.c_str() );
	  }//cimg
    else 
	  {
		cerr << "The file could not be read" << endl;
		exit(1);
	  }
    
    image_vector.resize( n_parameter_values );
    for( int p=0; p<n_parameter_values; ++p )
	  {
		const size_t zcells = Grid::dimensions==3 ? grid.cells()[2] : 1;
		image_vector[p] = Image( grid.cells()[0], grid.cells()[1], zcells, 1 );
		rescale_image( small_image[p], image_vector[p] );
	  }

    // Some output for convenience
    cout << "Loaded parameter field of dimension "
	 << image_vector[0].dimx() << " x " 
	 << image_vector[0].dimy() << " x " 
	 << image_vector[0].dimz() << endl;
    cout << "Number of pixels: "
	 << image_vector[0].size() << endl;

    // Store the cell dimensions to enable fast calculation of
    // parameter values from position
    cell_length[0] = extent[0]/(double)image_vector[0].dimx();
    cell_length[1] = extent[1]/(double)image_vector[0].dimy();

    if(dimensions == 3)
      cell_length[2]=extent[2]/(double)image_vector[0].dimz();

  }

  //! Return the parameter value at the coordinates given by
  //! position
  double operator() (const Vector & position, const int p = 0) const
  {
    assert(p<n_parameter_values);

    std::vector<size_t> x(dimensions);
    for(int d=0; d<dimensions; ++d)
      x[d] = (size_t)( position[d]/cell_length[d] );
    
    if(dimensions==3)
      return image_vector[p].atXYZV(x[0],x[1],x[2],0);
    else if(dimensions==2)
      return image_vector[p].atXYZV(x[0],x[1],0,0);
    else{
      assert(false);
    }
    return image_vector[p][0];
  }

  //! Return the parameter value for the cell corresponding
  //! to the grid index id
  const double & operator() (const size_t & id, const int p = 0) const
  {
    assert(p<n_parameter_values);

    assert(id < image_vector[p].size());
    return image_vector[p][id];
  }

  //! Return (for modification) the parameter value for the cell
  //! corresponding to the grid index id
  double & operator() ( const size_t & id, const int p = 0 )
  {
    assert(p<n_parameter_values);
    assert(id < image_vector[p].size());
    return image_vector[p][id];
  }

  inline int parameters()
  {
    return n_parameter_values;
  }

  //! Draw an ascii image of a transection of the parameter field of
  //! parameter p at cell coordinate z
  void debug_asci_output(int p = 0, unsigned int z = 0)
  {
    if( p > parameters() )
	  {
		std::cerr << "Error in SpatialParameters: Can not write ASCII output "
		  "for p > number of parameters!" << std::endl; 
		return ;
	  }

    size_t dimx = image_vector[p].dimx();
    size_t dimy = image_vector[p].dimy();

    for(size_t i = 0; i<dimy; ++i){
      for(size_t j = 0; j < dimx; ++j){
		cout << image_vector[p][j+(dimy-1-i)*dimx+z*dimx*dimy] << " ";
      }
      cout << endl;
    }
  }

  //! Save as image
  void save_as_image(const std::string filename, int p = 0)
  {
    if(p> parameters())
	  {
		std::cerr << "Error in SpatialParameters: Can not write ASCII output "
		  "for p > number of parameters!" << std::endl; 
		return ;
	  }
	
    Image copyimage( image_vector[p] );

    const size_t imagesize = copyimage.size();
    double min = std::numeric_limits<double>::max();
    double max = std::numeric_limits<double>::min();
    for(size_t i=0; i<imagesize; ++i)
	  {
		if(min > copyimage[i])
		  min = copyimage[i];
		if(max < copyimage[i])
		  max = copyimage[i];
	  }
  
    // color palette: 0..255
    int min_value = 1000;
    int max_value = 0;

    for(size_t i=0; i<imagesize; ++i)
      {
		// Rescaling values to adjust to color palette values!
		int value = ( copyimage[i] - min ) / ( max - min ) * 255;
		copyimage[i] = value;
		min_value = std::min( value, min_value );
		max_value = std::max( value, max_value );
      }

    cout << "Saving " << filename.c_str() << endl;
    cout << "imagesize = " << imagesize << endl;
    cout << "min_value = " << min_value << endl;
    cout << "max_value = " << max_value << endl;
	
    copyimage.save( filename.c_str() );
  }

  
};

#endif
